;DCF77 Clock with extended display (15 digits: time, date, weekday)
;Controller: AT90S2313 or ATTiny2313 @ 4.194304 MHz external crystal
;Fuses for ATTiny2313: low=0xDD, high=0xDB, ext=0xFF
;
;Assemble with avra (http://avra.sourceforge.net/) only, this file contains
;directives specific for this assembler!
;
;(C) Arne Rossius, http://www.ebps.de.vu
;
;DISPLAY:
;  - 14 LED displays, 7 segments + dp + 7 LEDs for weekday (15th "display")
;  - common cathode, segments connected through PNP transistors
;  - common cathode lines using 2x ULN2003 (ULN2803 + ULN2003 with weekday)
;  - shift register for common lines, first output = left display
;  - shift register outouts are active high for ULN2003/ULN2803 inputs
;  - display order: HH:MM:SS YYYY-MM-DD [Weekday] (or DD-MM-YYYY, set below)
;  - weekday display (optional): 7 LEDs, output a = Monday, ..., g = Sunday
;  - weekday matrix (optional): rotated column cathode 5x7 dot matrix display
;
;PORTS:
;  PB0-6 = Segments (a..g), active-low
;  PB7   = decimal point (active-low) *or* weekday matrix column 7
;  PD0-1 = weekday matrix columns 1 and 2
;  PD2   = DCF77 input
;  PD3   = "disable display" switch (open: off, closed to GND: on)
;  PD4   = weekday matrix column 3
;  PD5   = Data outout for shift register, weekday matrix column 5
;  PD6   = Clock output for shift register, weekday matrix column 6
;
;VERSION HISTORY:
;  v0.2  2007-01-10  Arne Rossius
;  v0.3  2007-02-04  Arne Rossius
;    * added switch to disable display
;  v0.4  2008-03-15  Arne Rossius
;    * weekday output on virtual display #15
;    * auto-increment date and weekday
;    * settings for inverted outputs and reverse date format
;  v0.5  2008-04-04
;    * display leading zero for day of the month
;  v0.6  2008-04-22  Arne Rossius
;    * BUGFIX: auto-increment date: load ZH:ZL instead of YH:ZL (narf!)
;    * added option to disable weekday display (slightly improves brightness)
;  v0.7  2008-07-25  Arne Rossius
;    * BUGFIX: added support for inverted common outputs (setting existed since
;              version 0.4, but didn't have an effect)
;  v0.8  2008-08-09  Arne Rossius
;    * BUGFIX: all segments were lit when "disable display" switch was open and
;              "InvertSegments" setting enabled
;    * changed name of "DisableWeekday" setting to "WeekdayLEDs" and reversed
;      meaning of 0 and 1 (previously 1 meant disabled, now 1 means enabled)
;    * changed method for display blanking when "disable display" switch open
;    * added support for 7x5 dot matrix weekday display (decimal points have to
;      be wired differently if this option is used)
;  v0.8.1  2008-11-13  Arne Rossius
;    * added support for AT90S1200 controller (except for weekday matrix)
;    * replaced month length table by month length calculation
;    * don't require weekday matrix files if weekday matrix not used

;microcontroller definition file
.equ	AT90S1200 = 0 ;use AT90S1200 controller (0 for AT90S2313 or ATTiny2313)
.if (AT90S1200 == 0)
	.include "../include/2313def.inc"
	.message "Assembling for 90S2313 or Tiny2313"
.else
	.include "../include/1200def.inc"
	.message "Assembling for 90S1200!"
.endif

;DCF77 receiver
.equ	DCFPIN = PIND
.equ	DCFInput = 2 ;don't change
.equ	DCFInvert = 1 ;inverted output (negative pulses), most receivers do this!

;"Disable Display" switch
.equ	SWPIN = PIND
.equ	SWInput = 3

;LED display
.equ	SEGMENT = PORTB
.equ	COMMON = PORTD
.equ	DATA = 5
.equ	CLOCK = 6
.equ	MATRIX1 = 0
.equ	MATRIX2 = 1
.equ	MATRIX3 = 4
.equ	MATRIX5 = 5
.equ	MATRIX6 = 6
.equ	MATRIX7 = 7
.equ	ReverseDateFormat = 0 ;set to 1 for DD-MM-YYYY date format
.equ	InvertCommon = 0 ;set to 1 for active-low common outputs (for PNP)
.equ	InvertSegments = 0 ;set to 1 for active-high segment outputs (NPN)
.equ	InvertMatrix = 0 ;set to 1 for active-high weekday matrix outputs (NPN)
.equ	WeekdayLEDs = 1 ;set to 1 to enable virtual digit 15 for 7 weekday LEDs
.equ	WeekdayMatrix = 0 ;set to 1 to enable 7x5 dot matrix weekday display

;needed for lpm instruction
.def	lpmresult = R0

;registers for DCF77 telegram
.def	dcf1 = R3
.def	dcf2 = R4
.def	dcf3 = R5
.def	dcf4 = R6
.def	dcf5 = R7

;time/date registers
.def	subsec = R8
.def	second = R9
.def	minute = R10
.def	hour = R11
.def	day = R12
.def	month = R13
.def	year = R14
.def	wkday = R15 ;(weekday, 1=MO 7=SU)

;general registers
.def	temp = R16
.def	temp2 = R17

;DCF77-specific registers
.def	dcfbits = R18 ;received bits
.def	dcftime = R19 ;1/64-seconds since last edge
.def	flags1 = R20
.def	flags2 = R21

;additional registers for multiplexing
.def	display = R22 ;active display [0..13]
.def	matrix = R23 ;weekday matrix data for PORTD

;interrupt backup registers
.def	backup1 = R24
.def	backup2 = R25
.def	backup3 = R26

;bit definitions for the flags1 register
.equ	ErrBitCount = 0 ;last error: wrong number of bits
.equ	ErrParity = 1 ;last error: parity error
.equ	ErrBitLength = 2 ;last error: wrong bit length
.equ	ErrStartBit = 3 ;last error: start bit was not 1
.equ	TimeOK = 4 ;time is OK
.equ	DateOK = 5 ;date is OK (reset at midnight)
.equ	StopDCF = 6 ;stop DCF77 reception until a new telegram starts
.equ	DCFdisabled = 7 ;DCF77 receiver disabled (reset when a new hour begins)
.equ	ErrMask = 1<<ErrBitLength | 1<<ErrParity | 1<<ErrBitCount | 1<<ErrStartBit ;used to reset errors

;bit definitions for the flags2 register
.equ	LastState = DCFInput ;last state of DCF77 input, must be the same bit as on the port!

.cseg

;*******************************************************************************
;*
;*   MAIN PROGRAM
;*
;*******************************************************************************

;---------------------------------------
; Interrupt vectors
;---------------------------------------

;Reset
.org 0x000
	rjmp	start
	
;Timer 0 overflow
.org OVF0addr
	rjmp	ovf0int

;;---------------------------------------
;; Length of every month in days
;;---------------------------------------
;
;months:
;	;length of months + 1, packed BCD
;	;    Jan   Feb   Mar   Apr   May   Jun   Jul   Aug   Sep   Oct   Nov   Dec
;	.db 0x32, 0x29, 0x32, 0x31, 0x32, 0x31, 0x32, 0x32, 0x31, 0x32, 0x31, 0x32 ;normal
;	.db 0x32, 0x30, 0x32, 0x31, 0x32, 0x31, 0x32, 0x32, 0x31, 0x32, 0x31, 0x32 ;leap year

;---------------------------------------
; Initialization
;---------------------------------------
	
start:
.if (AT90S1200 == 0)
	;stack pointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
.endif
	
	;disable analog comparator
	ldi	temp, 0x80
	out	ACSR, temp
	
	;I/O ports
	ldi	temp, 0xFF
	out	DDRB, temp
	out	PORTB, temp
	ldi	temp, 0x73
	out	DDRD, temp
	ldi	temp, 0x7F
	out	PORTD, temp
	
	;timer
	ldi	temp, 1<<TOIE0 ;overflow int on
	out	TIMSK, temp
	ldi	temp, 0x04 ;Clk/256 = 64Hz int
	out	TCCR0, temp
	
	;registers
	ldi	dcfbits, 0
	ldi	dcftime, 0
	ldi	flags1, 1<<StopDCF
	ldi	flags2, 0
	ldi	display, 0
	clr	wkday
	
	;enable interrupts
	sei

;---------------------------------------
; Main loop
;   modifies: temp, temp2, display, ZL, ZH, lpmresult
;---------------------------------------

;multiplex all the time and do *all* the action while one of the displays is enabled
main_loop:
	;enable next display
.if (InvertSegments == 0)	;display off
	ldi	temp2, 0xFF
.else
	ldi	temp2, 0x00
.endif
	out	SEGMENT, temp2
	cpi	display, 0	;select next display
	sbis	SWPIN, SWInput ;no display if switch is open
	breq	PC+3
.if ( InvertCommon == 0 )
	cbi	COMMON, DATA
	rjmp	PC+2
	sbi	COMMON, DATA
.else
	sbi	COMMON, DATA
	rjmp	PC+2
	cbi	COMMON, DATA
.endif
	cbi	COMMON, CLOCK
	sbi	COMMON, CLOCK
	out	COMMON, matrix ;display on (output new data)
	out	SEGMENT, temp
	
	;encode the next display
	inc	display
.if ( WeekdayLEDs == 0 )
	cpi	display, 14 ;return to display 0 after display 13 (no weekday LEDs)
.else
	cpi	display, 15 ;return to display 0 after display 14
.endif
	brlo	PC+2
	ldi	display, 0
	
	;set data for display
	cpi	display, 2
	brlo	disp_hr
	cpi	display, 4
	brlo	disp_min
	cpi	display, 6
	brlo	disp_sec
.if (ReverseDateFormat == 0)
	cpi	display, 8 ;YYYY-MM-DD format
	brlo	disp_20
	cpi	display, 10
	brlo	disp_yr
	cpi	display, 12
	brlo	disp_mon
	cpi	display, 14
	brlo	disp_day
.else
	cpi	display, 8 ;DD-MM-YYYY format
	brlo	disp_day
	cpi	display, 10
	brlo	disp_mon
	cpi	display, 12
	brlo	disp_20
	cpi	display, 14
	brlo	disp_yr
.endif
	mov	ZL, wkday	;weekday
.if (InvertSegments == 0)
	ldi	temp, 0xFE
	sbrs	flags1, TimeOK
	ldi	temp, 0xFF ;" "
.else
	ldi	temp, 0x01
	sbrs	flags1, TimeOK
	ldi	temp, 0x00 ;" "
.endif
disp_wkday_loop:
	cpi	ZL, 2
	brsh	PC+2
	rjmp	encode_end
	lsl	temp
.if (InvertSegments == 0)
	ori	temp, 0x01
.endif
	dec	ZL
	rjmp	disp_wkday_loop

disp_hr:			;hour (or # of bits received)
	mov	ZL, hour
	sbrc	flags1, TimeOK
	rjmp	display_end
	mov	ZL, dcfbits
	rcall	bin2bcd
	rjmp	display_end
disp_min:			;second (or last error)
	mov	ZL, minute
	sbrc	flags1, TimeOK
	rjmp	display_end
	ldi	ZL, 0xFE ;" -"
	sbrc	flags1, ErrBitCount
	ldi	ZL, 0xFA ;" C"
	sbrc	flags1, ErrBitLength
	ldi	ZL, 0xFB ;" L"
	sbrc	flags1, ErrParity
	ldi	ZL, 0xFC ;" P"
	sbrc	flags1, ErrStartBit
	ldi	ZL, 0xF5 ;" S"
	rjmp	display_end
disp_sec:			;second (or last received bit)
	mov	ZL, second
	sbrc	flags1, TimeOK
	rjmp	display_end
	ldi	ZL, 0xFE ;" -"
	cpi	dcfbits, 0
	breq	display_end
	sbrc	flags1, StopDCF
	rjmp	display_end
	ldi	ZL, 0xF0 ;" 0"
	sbrc	dcf5, 7
	ldi	ZL, 0xF1 ;" 1"
	rjmp	display_end

disp_20:			;century (20)
	ldi	ZL, 0x20
	sbrs	flags1, TimeOK
	ldi	ZL, 0xFF ;"  "
	rjmp	display_end
disp_yr:			;year
	mov	ZL, year
	sbrs	flags1, TimeOK
	ldi	ZL, 0xFF ;"  "
	rjmp	display_end
disp_mon:			;month
	mov	ZL, month
	sbrs	flags1, TimeOK
	ldi	ZL, 0xFF ;"  "
	rjmp	display_end
disp_day:
	mov	ZL, day		;day (or DCF77 input)
	sbrc	flags1, TimeOK
	rjmp	display_end
	ldi	ZL, 0xFD ;" _"
	sbrc	flags2, LastState
	ldi	ZL, 0xFE ;" -"
	rjmp	display_end
display_end:

	;suppress leading zero
	cpi	display, 0
	brne	PC+4
	cpi	ZL, 0x10
	brsh	PC+2
	ldi	ZL, 0xF0 ;15 = space
	
	;encode to 7-segment
	sbrs	display, 0 ;swap if 10's wanted
	swap	ZL
	andi	ZL, 0x0F
.if (AT90S1200 == 0)
	clr	ZH
	ldi	temp, LOW(charset*2)
	add	ZL, temp
	ldi	temp, HIGH(charset*2)
	adc	ZH, temp
	lpm
	mov	temp, lpmresult
.else
	ldi	temp, CHARSET_F
	cpi	ZL, 0x0
	brne	PC+2
	ldi	temp, CHARSET_0
	cpi	ZL, 0x1
	brne	PC+2
	ldi	temp, CHARSET_1
	cpi	ZL, 0x2
	brne	PC+2
	ldi	temp, CHARSET_2
	cpi	ZL, 0x3
	brne	PC+2
	ldi	temp, CHARSET_3
	cpi	ZL, 0x4
	brne	PC+2
	ldi	temp, CHARSET_4
	cpi	ZL, 0x5
	brne	PC+2
	ldi	temp, CHARSET_5
	cpi	ZL, 0x6
	brne	PC+2
	ldi	temp, CHARSET_6
	cpi	ZL, 0x7
	brne	PC+2
	ldi	temp, CHARSET_7
	cpi	ZL, 0x8
	brne	PC+2
	ldi	temp, CHARSET_8
	cpi	ZL, 0x9
	brne	PC+2
	ldi	temp, CHARSET_9
	cpi	ZL, 0xA
	brne	PC+2
	ldi	temp, CHARSET_A
	cpi	ZL, 0xB
	brne	PC+2
	ldi	temp, CHARSET_B
	cpi	ZL, 0xC
	brne	PC+2
	ldi	temp, CHARSET_C
	cpi	ZL, 0xD
	brne	PC+2
	ldi	temp, CHARSET_D
	cpi	ZL, 0xE
	brne	PC+2
	ldi	temp, CHARSET_E
.endif
	
	;enable dp after hours, minutes, year and month
	sbrs	flags1, TimeOK
	rjmp	dp_end
	cpi	display, 5
	breq	dp_end
.if (ReverseDateFormat == 0)
	cpi	display, 7
.else
	cpi	display, 11
.endif
	breq	dp_end
	cpi	display, 13
	breq	dp_end
	sbrc	display, 0
.if (InvertSegments == 0)
	andi	temp, ~0x80
.else
	ori	temp, 0x80
.endif
dp_end:
encode_end:

	;generate weekday matrix data
.if (InvertMatrix == 0)
	ldi	matrix, 0xFF
.else
	ldi	matrix, 1<<DCFInput | 1<<SWInput ;keep pullups on inputs enabled
.endif
	cpi	display, 9
	brlo	matrix_end
	cpi	display, 14
	brsh	matrix_end
	mov	ZL, wkday ;ZH:ZL = wkday * 5
	lsl	ZL
	lsl	ZL
	add	ZL, wkday
	clr	ZH
	ldi	temp2, LOW(weekdays*2-9)
	add	ZL, temp2
	ldi	temp2, HIGH(weekdays*2-9)
	adc	ZH, temp2
	ldi	temp2, 0
	add	ZL, display
	adc	ZH, temp2
.if (AT90S1200 == 0)
	lpm
.endif
	mov	matrix, lpmresult
	ori	matrix, 1<<DCFInput | 1<<SWInput ;keep pullups on inputs enabled
matrix_end:

	;add weekday matrix data for PB7
.if (WeekdayMatrix != 0)
.if (AT90S1200 != 0)
	.error "Weekday Matrix not supported with AT90S1200!"
.endif
.if (InvertMatrix == 0)
	ori	temp, 0x80
	sbrc	matrix, 7
	andi	temp, 0x7F
.else
	andi	temp, 0x7F
	sbrc	matrix, 7
	ori	temp, 0x80
.endif
.endif
	
	;wait for a while
	ldi	ZH, 2
	ldi	ZL, 0
multiplex_wait:
	dec	ZL
	brne	multiplex_wait
	dec	ZH
	brne	multiplex_wait
	
	rjmp	main_loop
	
;---------------------------------------
; 7-Segment encoding table
;---------------------------------------

charset:
	.equ	CHARSET_0 = 0x3F ; 0
	.equ	CHARSET_1 = 0x06 ; 1
	.equ	CHARSET_2 = 0x5B ; 2
	.equ	CHARSET_3 = 0x4F ; 3
	.equ	CHARSET_4 = 0x66 ; 4
	.equ	CHARSET_5 = 0x6D ; 5
	.equ	CHARSET_6 = 0x7D ; 6
	.equ	CHARSET_7 = 0x07 ; 7
	.equ	CHARSET_8 = 0x7F ; 8
	.equ	CHARSET_9 = 0x6F ; 9
	.equ	CHARSET_A = 0x39 ; C
	.equ	CHARSET_B = 0x38 ; L
	.equ	CHARSET_C = 0x73 ; P
	.equ	CHARSET_D = 0x08 ; _
	.equ	CHARSET_E = 0x40 ; -
	.equ	CHARSET_F = 0x00 ; (Space)
.if (InvertSegments == 0)
	.db ~CHARSET_0, ~CHARSET_1, ~CHARSET_2, ~CHARSET_3
	.db ~CHARSET_4, ~CHARSET_5, ~CHARSET_6, ~CHARSET_7
	.db ~CHARSET_8, ~CHARSET_9, ~CHARSET_A, ~CHARSET_B
	.db ~CHARSET_C, ~CHARSET_D, ~CHARSET_E, ~CHARSET_F
.else
	.db CHARSET_0, CHARSET_1, CHARSET_2, CHARSET_3
	.db CHARSET_4, CHARSET_5, CHARSET_6, CHARSET_7
	.db CHARSET_8, CHARSET_9, CHARSET_A, CHARSET_B
	.db CHARSET_C, CHARSET_D, CHARSET_E, CHARSET_F
.endif

;---------------------------------------
; Weekday Matrix encoding table
;---------------------------------------

weekdays:
.if (WeekdayMatrix != 0)
.if (InvertMatrix == 0)
	.include "weekdays/german-uc.asm"
.else
	.include "weekdays/german-uc-inv.asm"
.endif
.endif

;*******************************************************************************
;*
;*   DCF77 subroutines
;*
;*******************************************************************************

;---------------------------------------
; Timer 0 overflow int (64Hz)
;   modifies: dcftime, dcfbits, flags1, flags2, dcf1, dcf2, dcf3, dcf4, dcf5,
;             subsec, second, minute, hour, day, wkday, month, year
;---------------------------------------

ovf0int:
	;save registers
.if (AT90S1200 == 0)
	push	temp
	push	temp2
	in	temp, SREG
	push	temp
.else
	mov	backup1, temp
	mov	backup2, temp2
	in	backup3, SREG
.endif
	
	;don't evaluate DCF77 signal when reception is disabled
	sbrc	flags1, DCFdisabled
	rjmp	dcf_end
	
	inc	dcftime
	
	;compare DCF77 input to last saved state
	in	temp, DCFPIN
	eor	temp, flags2
	andi	temp, 1<<LastState
	.if ( DCFInvert )
		.message "DCF77 Signal: INVERTED"
		brne	dcf_end ;no change
	.else
		.message "DCF77 Signal: normal"
		breq	dcf_end ;no change
	.endif
	
	;take appropriate action for pos/neg edges
	sbrs	flags2, LastState
	rjmp	dcf_posedge
	
		; < negative edge ( 1 -> 0 ) >
	
	andi	flags2, ~(1<<LastState)
	
	;exit if an error was detected before
	sbrc	flags1, StopDCF
	rjmp	dcf_end
	 
	;increment bit number, exit if more than 59 bits were already received
	inc	dcfbits
	cpi	dcfbits, 60
	brsh	PC+2
	rjmp	dcf_evalbit
	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitCount | 1<<StopDCF
	rjmp	dcf_end
	
		; < positive edge ( 0 -> 1 ) >
dcf_posedge:
	ori	flags2, 1<<LastState
	
	;check if time was >1.2s (start of new telegram)
	cpi	dcftime, 77
	ldi	dcftime, 0
	brlo	dcf_end
	
	;end of telegram / new telegram
	andi	flags1, ~ErrMask & ~(1<<StopDCF)
	cpi	dcfbits, 59
	ldi	dcfbits, 0
	brne	PC+2
	rjmp	dcf_eval
	ori	flags1, 1<<ErrBitCount
dcf_end:

;---------------------------------------
; increment time (fallback quartz clock)
;---------------------------------------

	;increment 1/64-seconds
	inc	subsec
	ldi	temp, 64
	cp	subsec, temp
	brlo	inctime_end
	
	;increment seconds
	clr	subsec
	mov	temp, second
	rcall	bcdinc
	mov	second, temp
	cpi	temp, 0x60
	brlo	inctime_end
	
	;increment minutes
	clr	second
	mov	temp, minute
	rcall	bcdinc
	mov	minute, temp
	cpi	temp, 0x04 ;disable DCF77 reception if nothing received for 4 minutes
	brne	inctime_nodisable
	sbrc	flags1, TimeOK
	ori	flags1, 1<<DCFdisabled
inctime_nodisable:
	cpi	temp, 0x60
	brlo	inctime_end

	;increment hours
	clr	minute
	andi	flags1, ~(1<<DCFdisabled) ;re-enable DCF77 reception
	mov	temp, hour
	rcall	bcdinc
	mov	hour, temp
	cpi	temp, 0x24
	brlo	inctime_end
	
	;increment weekday and days
	clr	hour
	mov	temp, wkday
	inc	temp
	cpi	temp, 8
	brlo	PC+2
	ldi	temp, 1
	mov	wkday, temp
;	ldi	ZL, LOW(months*2-1) ;get length of current month
;	ldi	ZH, HIGH(months*2-1)
;	sbrc	year, 0
;	rjmp	noleap
;	sbrc	year, 1
;	rjmp	noleap
;	adiw	ZH:ZL, 12
;noleap:
;	mov	temp, month
;	rcall	bcd2bin
;	add	ZL, temp
;	ldi	temp, 0
;	adc	ZH, temp
;	lpm
	ldi	temp2, 0x30 ;calculate length of month in days
	mov	temp, month
	cpi	temp, 0x08
	brlo	PC+2
	inc	temp
	sbrc	temp, 0
	ldi	temp2, 0x31
	cpi	temp, 0x02
	brne	monthlength_end
	ldi	temp2, 0x28
	mov	temp, year
	andi	temp, 0x03
	brne	monthlength_end
	ldi	temp2, 0x29
monthlength_end:
	mov	temp, day ;increment day
	rcall	bcdinc
	mov	day, temp
	cp	temp, temp2 ;temp2 is length of current month
	brlo	inctime_end
	
	;increment month
	clr	day
	inc	day
	mov	temp, month
	rcall	bcdinc
	mov	month, temp
	cpi	temp, 0x13
	brlo	inctime_end
	
	;increment year
	clr	month
	inc	month
	mov	temp, year
	rcall	bcdinc
	mov	year, temp
inctime_end:

.if (AT90S1200 == 0)
	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
.else
	mov	temp, backup1
	mov	temp2, backup2
	out	SREG, backup3
.endif
	reti

;---------------------------------------
; DCF77 reception (continued)
;---------------------------------------

		; < negative edge - cont'd >
	
	;check bit length (optional)
dcf_evalbit:
	cpi	dcftime, 5 ;bit is 0: 5 <= dcftime <= 7
	brlo	dcf_errbit
	cpi	dcftime, 8
	brlo	dcf_clearbit
	cpi	dcftime, 11 ;bit is 1: 11 <= dcftime <= 13
	brlo	dcf_errbit
	cpi	dcftime, 14
	brlo	dcf_setbit
	
	;received bit is erroneous
dcf_errbit:
	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitLength | 1<<StopDCF
	rjmp	dcf_end
	
	;save bit
dcf_clearbit:
	clc ;clear bit
	rjmp	dcf_savebit
dcf_setbit:
	sec ;set bit
dcf_savebit:
	ror	dcf5
	ror	dcf4
	ror	dcf3
	ror	dcf2
	ror	dcf1
	rjmp	dcf_end
	
;---------------------------------------
; evaluate DCF77 telegram
;---------------------------------------
	
dcf_eval:
	;check date parity
	ldi	temp2, 0
	mov	temp, dcf3
	andi	temp, 0xFE
	rcall	parity
	mov	temp, dcf4
	rcall	parity
	mov	temp, dcf5
	andi	temp, 0x7F
	rcall	parity
	mov	temp, dcf5
	lsl	temp ;Bit7 -> C
	rol	temp ;C -> Bit0
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check hour parity
	ldi	temp2, 0
	mov	temp, dcf2
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf3
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check minute parity
	ldi	temp2, 0
	mov	temp, dcf1
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf2
	andi	temp, 0x01
	rcall	parity
	mov	temp, dcf2
	lsr	temp
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror

	;save minutes
	lsr	dcf2
	ror	dcf1
	lsr	dcf1
	mov	minute, dcf1
	
	;save hours
	lsr	dcf2
	mov	hour, dcf2
	
	;save day
	lsl	dcf3 ;align day
	rol	dcf4 ;align month[5] weekday[3]
	rol	dcf5 ;align year
	lsr	dcf3
	lsr	dcf3
	mov	day, dcf3
	
	;save weekday
	mov	wkday, dcf4
	ldi	temp, 0x07
	and	wkday, temp
	
	;save month
	lsr	dcf4
	lsr	dcf4
	lsr	dcf4
	mov	month, dcf4
	
	;save year
	mov	year, dcf5
	
	;set time & date ok, disable DCF77 reception
	clr	subsec
	clr	second
	ori	flags1, 1<<TimeOK | 1<<DateOK | 1<<DCFdisabled
	rjmp	dcf_end
	
dcf_parityerror:
	;set parity error
	ori	flags1, 1<<ErrParity
	rjmp	dcf_end

;*******************************************************************************
;*
;*   GENERAL ARITHMETIC FUNCTIONS
;*
;*******************************************************************************

;---------------------------------------
; calculate parity and add up
;   in: temp
;   out: temp2 (add to existing value)
;---------------------------------------

parity:
	lsr	temp
	brcc	PC+2
	inc	temp2
	cpi	temp, 0
	brne	parity
	andi	temp2, 0x01
	ret
	
;---------------------------------------
; increment packed bcd number
;   in: temp
;   out: temp
;---------------------------------------

bcdinc:
	subi	temp, -7 ;increment and add 6
	brhc	PC+2 ;half-carry is inverted when using "subi rh, -k" instead of add!
	subi	temp, 6
	ret
	
;---------------------------------------
; convert binary to packed BCD
;   in: ZL
;   out: ZL
;   modifies: temp
;---------------------------------------

bin2bcd:
	ldi	temp, 0
bin2bcd_loop:
	subi	temp, -0x10
	subi	ZL, 10
	brcc	bin2bcd_loop
	subi	temp, 0x10
	subi	ZL, -10
	add	ZL, temp
	ret
	
;---------------------------------------
; convert packed BCD to binary
;   in: temp
;   out: temp
;   modifies: temp2
;---------------------------------------

bcd2bin:
	mov	temp2, temp
	andi	temp, 0x0F
	swap	temp2
	andi	temp2, 0x0F
	lsl	temp2
	add	temp, temp2 ;temp += temp2*2
	lsl	temp2
	lsl	temp2
	add	temp, temp2 ;temp += temp2*8
	ret
